package com.rwtema.extrautils2.blocks;

import com.rwtema.extrautils2.backend.XUBlockConnectedTextureBase;
import com.rwtema.extrautils2.utils.LogHelper;
import net.minecraft.block.material.Material;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.SharedMonsterAttributes;
import net.minecraft.entity.ai.attributes.AttributeModifier;
import net.minecraft.entity.ai.attributes.IAttribute;
import net.minecraft.entity.ai.attributes.IAttributeInstance;
import net.minecraft.world.WorldServer;

import java.util.List;
import java.util.UUID;

abstract class BlockCursedEarth extends XUBlockConnectedTextureBase {
    public BlockCursedEarth(Material materialIn) {
        super(materialIn);
    }

    static final UUID uuid = UUID.fromString("E53E0344-EA5E-4F71-98F6-40791198D8FE");

    public static boolean spawnMobAsCursed(Entity mob) {
        if(!mob.world.isRemote && ((WorldServer)mob.world).getEntityTracker().trackedEntityHashTable.containsItem(mob.getEntityId())) {
            LogHelper.info("LaoAFix: already tracked entity " + mob.getClass() + "@" + mob.getEntityId());
            return false;
        }

        mob.forceSpawn = true;
        if (mob instanceof EntityLivingBase) {
            mob.getEntityData().setInteger("CursedEarth", 60);
            EntityLivingBase living = (EntityLivingBase)mob;
            applyAttribute(living, SharedMonsterAttributes.ATTACK_DAMAGE, new AttributeModifier(uuid, "CursedEarth", 1.5D, 1));
            applyAttribute(living, SharedMonsterAttributes.MOVEMENT_SPEED, new AttributeModifier(uuid, "CursedEarth", 1.2D, 1));
            IAttributeInstance inst = living.getAttributeMap().getAttributeInstanceByName("zombie.spawnReinforcements");
            if (inst != null)
                inst.setBaseValue(0.0D);
        }

        if (!mob.world.spawnEntity(mob))
            return false;
        if (mob.isBeingRidden()) {
            List<Entity> passengers = mob.getPassengers();
            for (int i = 0; i < passengers.size(); i++) {
                spawnMobAsCursed(passengers.get(i));
            }
        }
        return true;
    }

    private static void applyAttribute(EntityLivingBase mob, IAttribute attackDamage, AttributeModifier modifier) {
        IAttributeInstance instance = mob.getEntityAttribute(attackDamage);
        if (instance != null)
            instance.applyModifier(modifier);
    }
}

