package com.rwtema.extrautils2.gui.backend;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.LinkedHashMultimap;
import com.rwtema.extrautils2.ExtraUtils2;
import com.rwtema.extrautils2.backend.ISidedFunction;
import com.rwtema.extrautils2.backend.model.XUBlockState;
import com.rwtema.extrautils2.compatibility.CompatHelper;
import com.rwtema.extrautils2.compatibility.StackHelper;
import com.rwtema.extrautils2.network.*;
import com.rwtema.extrautils2.network.NetworkHandler.XUPacket;
import com.rwtema.extrautils2.tile.XUTile;
import com.rwtema.extrautils2.utils.ItemStackNonNull;
import com.rwtema.extrautils2.utils.Lang;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufUtil;
import io.netty.buffer.Unpooled;
import net.minecraft.client.Minecraft;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.entity.player.InventoryPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.inventory.*;
import net.minecraft.item.ItemStack;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.text.TextComponentString;
import net.minecraft.util.text.translation.I18n;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;
import net.minecraftforge.items.ItemHandlerHelper;
import org.apache.commons.lang3.tuple.Pair;

import javax.annotation.Nonnull;
import java.util.*;

public abstract class DynamicContainer extends Container {
	public static final int playerInvWidth = 162;

	public static final int centerX = 85;

	public static final int centerSlotX = 76;

	public static final int playerInvHeight = 95;

	public static final ISidedFunction<String, Integer> STRING_WIDTH_FUNCTION = new ISidedFunction<String, Integer>() {
		@SideOnly(Side.SERVER)
		public Integer applyServer(String input) { return (input != null) ? input.length() : 0; }

		@SideOnly(Side.CLIENT)
		public Integer applyClient(String input) { return (input != null) ?
						(Minecraft.getMinecraft()).fontRenderer.getStringWidth(input) : 0; }
	};

	public static final ISidedFunction<Pair<String, Integer>, Integer> STRING_HEIGHTS =
					new ISidedFunction<Pair<String, Integer>, Integer>() {
						public Integer applyServer(Pair<String, Integer> input) {
							int w = 0;
							String[] split = input.getKey().split("\n");
							for (String s : split) {
								float sw = DynamicContainer.STRING_WIDTH_FUNCTION.applyServer(s);
								w += MathHelper.ceil(sw / w);
							}
							return w * 9;
						}

						public Integer applyClient(Pair<String, Integer> input) {
							List<String> strings =
											(Minecraft.getMinecraft()).fontRenderer.listFormattedStringToWidth(input.getKey(), input.getValue());
							return strings.size() * 9;
						}
					};

	public static final ResourceLocation texBackground = new ResourceLocation("extrautils2", "textures/gui_base.png");

	public static final ResourceLocation texBackgroundBlack = new ResourceLocation("extrautils2", "textures" +
					"/gui_base_black.png");

	public static final ResourceLocation texBackgroundIndentation = new ResourceLocation("extrautils2", "textures" +
					"/gui_base_indent.png");

	public static final ResourceLocation texBackgroundBorder = new ResourceLocation("extrautils2", "textures" +
					"/gui_base_invisible.png");

	private static final ItemStack genericItemStack = new ItemStack(Blocks.COBBLESTONE, 65);

	public final HashSet<Slot> playerSlots = new HashSet<>();

	public int playerSlotsStart = -1;

	public LinkedList<EntityPlayerMP> entityPlayerMPs = new LinkedList<>();

	public int width = 176;

	public int height = 166;

	public boolean changesOnly = false;

	public boolean isClient = false;

	ByteBuf previousPacket = Unpooled.buffer();

	boolean fixed = false;

	private final LinkedHashMultimap<DynamicWindow, IWidget> windowWidgets = LinkedHashMultimap.create();

	private final List<IWidget> widgets = new ArrayList<>();

	private final List<IWidgetKeyInput> widgetKeyInputs = new ArrayList<>();

	private final List<IWidgetMouseInput> widgetMouseInputs = new ArrayList<>();

	private final List<IWidgetClientNetwork> widgetReceivers = new ArrayList<>();

	private final List<IWidgetServerNetwork> widgetNetworks = new ArrayList<>();

	private final List<IWidgetClientTick> widgetClientTick = new ArrayList<>();

	private final HashMap<IWidget, DynamicWindow> windowOwner = new HashMap<>();

	private final HashSet<DynamicWindow> windows = new HashSet<>();

	public List<IWidgetClientTick> getWidgetClientTick() { return this.widgetClientTick; }

	public HashMap<IWidget, DynamicWindow> getWindowOwner() { return this.windowOwner; }

	public LinkedHashMultimap<DynamicWindow, IWidget> getWindowWidgets() { return this.windowWidgets; }

	public List<IWidgetKeyInput> getWidgetKeyInputs() { return ImmutableList.copyOf(this.widgetKeyInputs); }

	public List<IWidgetMouseInput> getWidgetMouseInputs() { return ImmutableList.copyOf(this.widgetMouseInputs); }

	public List<IWidget> getWidgets() { return ImmutableList.copyOf(this.widgets); }

	void addSlot(Slot slot) { addSlotToContainer(slot); }

	public void addListener(IContainerListener listener) {
		if (listener instanceof EntityPlayerMP) this.entityPlayerMPs.add((EntityPlayerMP) listener);
		this.changesOnly = false;
		super.addListener(listener);
	}

	public void removeListener(@Nonnull IContainerListener listener) {
		if (listener instanceof EntityPlayerMP) this.entityPlayerMPs.remove(listener);
		super.removeListener(listener);
	}

	public void detectAndSendChanges() {
		if (this.isClient) return;
		super.detectAndSendChanges();
		ByteBuf buffer = Unpooled.buffer();
		XUPacketBuffer packetBuffer = new XUPacketBuffer(buffer);
		for (IWidgetServerNetwork widget : this.widgetNetworks)
			widget.addToDescription(packetBuffer);
		this.previousPacket.setIndex(0, this.previousPacket.writerIndex());
		if (ByteBufUtil.equals(buffer, this.previousPacket)) return;
		this.previousPacket = Unpooled.copiedBuffer(buffer);
		for (EntityPlayerMP player : this.entityPlayerMPs)
			NetworkHandler.sendPacketToPlayer(new PacketGUIData(this.windowId, packetBuffer), player);
	}

	protected void validate() {
		boolean flag = false;
		for (IWidget widget : this.widgets) {
			widget.addToContainer(this);
			if (!flag && widget instanceof ITransferPriority) flag = true;
		}
		this.fixed = true;
		if (flag) {
			this.inventorySlots.sort((o1, o2) -> {
				int p1 = (o1 instanceof ITransferPriority) ? ((ITransferPriority) o1).getTransferPriority() : 0;
				int p2 = (o2 instanceof ITransferPriority) ? ((ITransferPriority) o2).getTransferPriority() : 0;
				int compare = -Integer.compare(p1, p2);
				if (compare != 0) return compare;
				return Integer.compare(o1.slotNumber, o2.slotNumber);
			});
			for (int i = 0; i < this.inventorySlots.size(); i++) {
				Slot inventorySlot = this.inventorySlots.get(i);
				inventorySlot.slotNumber = i;
			}
		}
	}

	public void addPlayerSlotsToBottom(IInventory inventory) { addPlayerSlots(inventory, (this.width - 162) / 2,
					this.height - 95); }

	public void crop() { crop(4); }

	public void crop(int border) {
		int maxX = 18;
		int maxY = 18;
		for (IWidget widget : this.windowWidgets.get(null)) {
			maxX = Math.max(maxX, widget.getX() + widget.getW());
			maxY = Math.max(maxY, widget.getY() + widget.getH());
		}
		this.width = maxX + border;
		this.height = maxY + border;
	}

	public void cropForPlayerSlots() {
		crop(4);
		if (this.width < 170) this.width = 170;
	}

	public void cropAndAddPlayerSlots(InventoryPlayer inventory) {
		crop(4);
		this.height += 95;
		if (this.width < 170) this.width = 170;
		addPlayerSlotsToBottom(inventory);
	}

	public void addPlayerSlots(IInventory inventory, int x, int y) {
		this.playerSlotsStart = 0;
		for (IWidget widget : this.widgets) {
			if (widget instanceof Slot) this.playerSlotsStart++;
		}
		addWidget(new WidgetTextTranslate(x, y, inventory.getName(), 162));
		for (int j = 0; j < 3; j++) {
			for (int k = 0; k < 9; k++) {
				WidgetSlot w = new WidgetSlot(inventory, k + j * 9 + 9, x + k * 18, y + 14 + j * 18);
				this.playerSlots.add(w);
				addWidget(w);
			}
		}
		for (int j = 0; j < 9; j++) {
			WidgetSlot w = new WidgetSlot(inventory, j, x + j * 18, y + 14 + 58);
			this.playerSlots.add(w);
			addWidget(w);
		}
	}

	public void addWidget(IWidget w) { addWidget(w, null); }

	public void addWidgetClientTick(IWidgetClientTick w) { this.widgetClientTick.add(w); }

	public void onContainerClosed(EntityPlayer playerIn) {
		for (IWidget widget : this.widgets)
			widget.onContainerClosed(this, playerIn);
	}

	public void addWidget(IWidget w, DynamicWindow dynamicWindow) {
		if (this.fixed) throw new IllegalStateException();
		this.widgets.add(w);
		if (dynamicWindow != null) this.windows.add(dynamicWindow);
		this.windowWidgets.put(dynamicWindow, w);
		this.windowOwner.put(w, dynamicWindow);
		if (w instanceof IWidgetKeyInput) this.widgetKeyInputs.add((IWidgetKeyInput) w);
		if (w instanceof IWidgetMouseInput) this.widgetMouseInputs.add((IWidgetMouseInput) w);
		if (w instanceof IWidgetClientNetwork) this.widgetReceivers.add((IWidgetClientNetwork) w);
		if (w instanceof IWidgetServerNetwork) this.widgetNetworks.add((IWidgetServerNetwork) w);
		if (w instanceof IWidgetClientTick) this.widgetClientTick.add((IWidgetClientTick) w);
		if (w instanceof IAdditionalWidgets) for (IWidget iWidget : ((IAdditionalWidgets) w).getAdditionalWidgets())
			addWidget(iWidget, dynamicWindow);
	}

	@ItemStackNonNull
	public ItemStack transferStackInSlot(EntityPlayer player, int par2) {
		player.sendMessage(new TextComponentString("抱歉,因为懒得计算ID,我直接禁了shift"));
		return ItemStack.EMPTY;
	}

	protected boolean mergeItemStack(ItemStack stack, int startIndex, int endIndex, boolean reverseDirection) {
		boolean flag = false;
		if (stack.isStackable()) {
			int i = reverseDirection ? (endIndex - 1) : startIndex;
			for (; (reverseDirection && i >= startIndex) || (!reverseDirection && i < endIndex); i += (reverseDirection ? -1 : 1)) {
				Slot slot = this.inventorySlots.get(i);
				if (isValidForMerging(slot)) {
					ItemStack currentSlotStack = slot.getStack();
					if (!StackHelper.isEmpty(currentSlotStack) && ItemHandlerHelper.canItemStacksStack(stack,
									currentSlotStack) && slot.isItemValid(stack)) {
						int j = StackHelper.getStacksize(currentSlotStack) + StackHelper.getStacksize(stack);
						int maxSize = Math.min(slot.getSlotStackLimit(), stack.getMaxStackSize());
						if (j <= maxSize) {
							StackHelper.setStackSize(stack, 0);
							slot.putStack(ItemHandlerHelper.copyStackWithSize(currentSlotStack, j));
							slot.onSlotChanged();
							flag = true;
							break;
						}
						if (maxSize - StackHelper.getStacksize(currentSlotStack) > 0) {
							StackHelper.decrease(stack, maxSize - StackHelper.getStacksize(currentSlotStack));
							slot.putStack(ItemHandlerHelper.copyStackWithSize(currentSlotStack, maxSize));
							slot.onSlotChanged();
							flag = true;
						}
					}
				}
			}
		}
		if (StackHelper.isEmpty(stack)) return flag;
		int i = reverseDirection ? (endIndex - 1) : startIndex;
		for (; (reverseDirection && i >= startIndex) || (!reverseDirection && i < endIndex); i += (reverseDirection ?
						-1 : 1)) {
			Slot slot = this.inventorySlots.get(i);
			if (isValidForMerging(slot))
				if (!slot.getHasStack() && StackHelper.isEmpty(slot.getStack()) && slot.isItemValid(stack)) {
					if (StackHelper.getStacksize(stack) > slot.getSlotStackLimit()) {
						int toRemove = Math.min(slot.getSlotStackLimit(), StackHelper.getStacksize(stack));
						ItemStack otherStack = stack.copy();
						StackHelper.setStackSize(otherStack, toRemove);
						StackHelper.decrease(stack, toRemove);
						slot.putStack(otherStack);
					} else {
						slot.putStack(StackHelper.safeCopy(stack));
						StackHelper.setStackSize(stack, 0);
					}
					slot.onSlotChanged();
					flag = true;
					break;
				}
		}
		return flag;
	}

	public boolean isValidForMerging(Slot slot) { return (!(slot instanceof WidgetCraftingMatrix.WidgetSlotIngredients) && !(slot instanceof WidgetCraftingMatrix.WidgetSlotCrafting)); }

	public int getStringWidth(String text) { return ExtraUtils2.proxy.apply(STRING_WIDTH_FUNCTION, text); }

	public void addTitle(String name) { addTitle(Lang.translate(name), false); }

	public void addTitle(XUTile tile) { addTitle(tile.getBlockState()); }

	public void addTitle(XUBlockState blockState) { addTitle((new ItemStack(blockState.getBlock(), 1,
					blockState.dropMeta)).getDisplayName()); }

	public void addTitle(String name, boolean translate) {
		WidgetText e;
		if (translate) {
			e = new WidgetTextTranslate(5, 5, name, getStringWidth(I18n.translateToLocal(name)));
		} else {
			e = new WidgetText(5, 5, name, getStringWidth(name));
		}
		addWidget(e);
	}

	public ItemStack slotClick(int slotId, int clickedButton, ClickType clickTypeIn, EntityPlayer playerIn) {
		if (slotId >= 0 && slotId < this.inventorySlots.size()) {
			Slot slot = this.inventorySlots.get(slotId);
			if (slot instanceof ISlotClick)
				return ((ISlotClick) slot).slotClick(this, slotId, clickedButton, clickTypeIn, playerIn);
		}
		return super.slotClick(slotId, clickedButton, clickTypeIn, playerIn);
	}

	@SideOnly(Side.CLIENT)
	public void sendInputPacket(IWidgetClientNetwork widget, XUPacketBuffer buffer) {
		int i = this.widgetReceivers.indexOf(widget);
		if (i < 0) throw new RuntimeException("Unable to find widget");
		NetworkHandler.sendPacketToServer(new PacketGUIInput(this.windowId, i, buffer));
	}

	public HashSet<DynamicWindow> getWindows() { return this.windows; }

	public void onSlotChanged(int index) { this.inventoryItemStacks.set(index, genericItemStack); }

	@SideOnly(Side.CLIENT)
	public void loadGuiDimensions(DynamicGui dynamicGui) { dynamicGui.setWidthHeight(this.width, this.height); }

	@SideOnly(Side.CLIENT)
	public boolean drawBackgroundOverride(DynamicGui gui) { return false; }

	@XUPacket
	public static class PacketGUIInput extends XUPacketClientToServer {
		private int windowId;

		private int widgetId;

		private XUPacketBuffer packetBuffer;

		private EntityPlayer player;

		public PacketGUIInput() {}

		public PacketGUIInput(int windowId, int widgetId, XUPacketBuffer packetBuffer) {
			this.windowId = windowId;
			this.widgetId = widgetId;
			this.packetBuffer = packetBuffer;
		}

		public void writeData() throws Exception {
			writeInt(this.windowId);
			writeShort(this.widgetId);
			if (this.packetBuffer == null) {
				writeInt(0);
			} else {
				writePacketBuffer(this.packetBuffer);
			}
		}

		public void readData(EntityPlayer player) {
			this.player = player;
			this.windowId = readInt();
			this.widgetId = readUnsignedShort();
			this.packetBuffer = readPacketBuffer();
		}

		public Runnable doStuffServer() {
			return () -> {
				Container openContainer = PacketGUIInput.this.player.openContainer;
				if (PacketGUIInput.this.windowId == 0 || openContainer.windowId != PacketGUIInput.this.windowId || !openContainer.canInteractWith(PacketGUIInput.this.player))
					return;
				DynamicContainer dynamicContainer = (DynamicContainer) openContainer;
				List<IWidgetClientNetwork> receivers = dynamicContainer.widgetReceivers;
				if (PacketGUIInput.this.widgetId < 0 || PacketGUIInput.this.widgetId >= receivers.size())
					return;
				receivers.get(PacketGUIInput.this.widgetId).receiveClientPacket(PacketGUIInput.this.packetBuffer);
			};
		}
	}

	@XUPacket
	public static class PacketSetGhost extends XUPacketClientToServer {
		private int windowId;

		private EntityPlayer player;

		private int slot;

		private ItemStack stack;

		public PacketSetGhost(int windowId, int slot, ItemStack stack) {
			this.windowId = windowId;
			this.slot = slot;
			this.stack = stack;
		}

		public PacketSetGhost() {}

		public void readData(EntityPlayer player) {
			this.player = player;
			this.windowId = readInt();
			this.slot = readInt();
			this.stack = readItemStack();
		}

		public void writeData() {
			writeInt(this.windowId);
			writeInt(this.slot);
			writeItemStack(this.stack);
		}

		public Runnable doStuffServer() {
			return () -> {
				Container openContainer = PacketSetGhost.this.player.openContainer;
				if (PacketSetGhost.this.windowId == 0 || openContainer.windowId != PacketSetGhost.this.windowId || !openContainer.canInteractWith(PacketSetGhost.this.player))
					return;
				DynamicContainer dynamicContainer = (DynamicContainer) openContainer;
				List<IWidget> widgets1 = dynamicContainer.widgets;
				IWidget iWidget = widgets1.get(PacketSetGhost.this.slot);
				if (iWidget instanceof WidgetSlotGhost)
					((WidgetSlotGhost) iWidget).putStack(PacketSetGhost.this.stack);
			};
		}
	}

	@XUPacket
	public static class PacketGUIData extends XUPacketServerToClient {
		private int windowId;

		private XUPacketBuffer packetBuffer;

		public PacketGUIData() {}

		public PacketGUIData(int windowId, XUPacketBuffer packetBuffer) {
			this.windowId = windowId;
			this.packetBuffer = packetBuffer;
		}

		public void writeData() {
			writeInt(this.windowId);
			writePacketBuffer(this.packetBuffer);
		}

		public void readData(EntityPlayer player) {
			this.windowId = readInt();
			this.packetBuffer = readPacketBuffer();
		}

		@SideOnly(Side.CLIENT)
		public Runnable doStuffClient() {
			return () -> {
				Container openContainer = (Minecraft.getMinecraft()).player.openContainer;
				if (PacketGUIData.this.windowId == 0 || openContainer.windowId != PacketGUIData.this.windowId)
					return;
				DynamicContainer dynamicContainer = (DynamicContainer) openContainer;
				for (IWidgetServerNetwork widget : dynamicContainer.widgetNetworks)
					widget.handleDescriptionPacket(PacketGUIData.this.packetBuffer);
			};
		}
	}
}
